from .formatters import FloatFormatter, ExponentialFormatter
from .table import Table
from .errors import FixError, RelativeError, AbsoluteError
from .columns import SerialNumberColumn, TextColumn, DataColumn

__all__ = ["Table",
           "FloatFormatter",
           "ExponentialFormatter",
           "FixError",
           "AbsoluteError",
           "RelativeError",
           "SerialNumberColumn",
           "TextColumn",
           "DataColumn"]



from abc import ABC, abstractmethod
import numpy as np
from numpy.typing import NDArray
from .formatters import Formatter, FloatFormatter, ExponentialFormatter
from .errors import Error
from .converter import convertUnitToLateX
from astropy.units import Quantity


class Column(ABC):

    def __init__(self, header: str):
        self._header = header

    def getHeader(self):
        return self._header

    @abstractmethod
    def getCell(self, row: int):
        pass


class TextColumn(Column):

    def __init__(self, header: str, texts: list):
        super(TextColumn, self).__init__(header)
        self.__texts = texts

    def getCell(self, row: int):
        return self.__texts[row]

    def __len__(self):
        return len(self.__texts)


class SerialNumberColumn(TextColumn):

    def __init__(self, header: str, rowCount: int):
        texts = []
        for i in range(rowCount):
            texts.append("{}.".format(i + 1))
        super(SerialNumberColumn, self).__init__(header, texts)


class DataColumn(Column):

    def __init__(self, header: str, data: NDArray[np.float64] | Quantity, error: Error = None, formatter: Formatter = None):
        super(DataColumn, self).__init__(header)

        if isinstance(data, Quantity):
            self.__data = data.value
            self._header = self._header + " [" + convertUnitToLateX(data.unit) + "]"
        elif isinstance(data, np.ndarray):
            self.__data = data
        else:
            raise Exception("Data must be of type numpy.ndarray or astropy.units.Quantity")
        dataOrder = np.max(np.ceil(np.abs(np.log10(self.__data))))

        if formatter is None:
            if dataOrder > 4:
                self.__formatter = ExponentialFormatter()
            else:
                self.__formatter = FloatFormatter()
        elif not isinstance(formatter, Formatter):
            raise Exception("The argument 'formatter' must be a subclass of latab.Formatter")
        else:
            self.__formatter = formatter

        if error is not None:
            if isinstance(error, Error):
                self.__errors = error.getErrors(self.__data)
            else:
                raise Exception("Error must be of type latab.Error")

    def getHeader(self):
        return self._header

    def getCell(self, row: int):
        if hasattr(self, "_DataColumn__errors"):
            return self.__formatter.format(self.__data[row], self.__errors[row])
        else:
            return self.__formatter.format(self.__data[row])

    def __len__(self):
        return len(self.__data)



from abc import ABC, abstractmethod
from astropy.units import Quantity
from numpy.typing import NDArray
import numpy as np


class Error(ABC):

    @abstractmethod
    def getErrors(self, data: NDArray[np.float64]):
        pass


class FixError(Error):

    def __init__(self, error: float | Quantity):
        if isinstance(error, Quantity):
            self.__error = error.value
        elif isinstance(error, float):
            self.__error = error
        else:
            raise Exception("Fix error must be of type float or astropy.units.Quantity")

    def getErrors(self, data: NDArray[np.float64]):
        return np.ones(len(data)) * self.__error


class AbsoluteError(Error):

    def __init__(self, errors: NDArray[np.float64] | Quantity):
        if isinstance(errors, Quantity):
            self.__errors = errors.value
        elif isinstance(errors, np.ndarray):
            self.__errors = errors
        else:
            raise Exception("Absolute error must be of type numpy.ndarray or astropy.units.Quantity")

    def getErrors(self, data: NDArray[np.float64]):
        # TODO length check
        return self.__errors


class RelativeError(Error):

    def __init__(self, error: float):
        if isinstance(error, float):
            self.__error = error
        else:
            raise Exception("Relative error must be of type float")

    def getErrors(self, data: NDArray[np.float64]):
        return data * self.__error



from astropy.units import UnitBase
import numpy as np


def convertUnitToLateX(unit: UnitBase):
    s = "$\mathrm{"
    bases = unit.bases
    powers = np.array(unit.powers)
    positivePowers = powers[powers > 0]
    negativePowers = powers[powers < 0]
    positivePowerCount = len(positivePowers)
    negativePowerCount = len(negativePowers)
    if positivePowerCount == 0:
        s += "1"
    else:
        for i in range(positivePowerCount):
            if i > 0:
                s += "\\cdot " + str(bases[i])
            else:
                s += str(bases[i])
            if (np.abs(positivePowers[i]) != 1):
                s += "^{" + str(np.abs(positivePowers[i])) + "}"
    if negativePowerCount > 0:
        s += "/"
        if negativePowerCount > 1:
            s += "("
        for i in range(negativePowerCount):
            if i > 0:
                s += "\\cdot " + str(bases[i + positivePowerCount])
            else:
                s += str(bases[i + positivePowerCount])
            if (np.abs(negativePowers[i]) != 1):
                s += "^{" + str(np.abs(negativePowers[i])) + "}"
        if negativePowerCount > 1:
            s += ")"
    return s + "}$"



from numpy import float64
from .columns import DataColumn, SerialNumberColumn, TextColumn
from astropy.units import Quantity
from numpy.typing import NDArray
from .formatters import Formatter
from .errors import Error


class Table():

    def __init__(self, caption: str = None):
        self.__columns = []
        self.__caption = caption

    def __checkRowCount(self, rowCount: int):
        if not hasattr(self, "_Table__rowCount"):
            self.__rowCount = rowCount
        elif rowCount != self.__rowCount:
            raise Exception("Columns have different lengths")

    def serialColumn(self, header: str, rowCount: int):
        self.__checkRowCount(rowCount)
        self.__columns.append(SerialNumberColumn(header, rowCount))
        return self

    def textColumn(self, header: str, texts: list):
        self.__checkRowCount(len(texts))
        self.__columns.append(TextColumn(header, texts))
        return self

    def dataColumn(self, header: str, data: NDArray[float64] | Quantity, error: Error = None, formatter: Formatter = None):
        self.__checkRowCount(len(data))
        self.__columns.append(DataColumn(header, data, error, formatter))
        return self

    def lines(self, tabLength: int = 4, separator: chr = '.'):
        lines = []
        lines.append("\\begin{table}")
        lines.append("\t\\centering".expandtabs(tabLength))
        lines.append(("\t\\begin{tabular}{|" + "c|" * len(self.__columns) + "} \\hline").expandtabs(tabLength))
        s = "\t\t".expandtabs(tabLength)
        for column in self.__columns:
            s += column.getHeader()
            s += " & "
        s = s[0:-2]
        s += "\\\\ \hline"
        lines.append(s)
        for i in range(self.__rowCount):
            s = "\t\t".expandtabs(tabLength)
            for column in self.__columns:
                cell = column.getCell(i)
                if isinstance(column, DataColumn) and separator != '.':
                    s += cell.replace(".", separator)
                else:
                    s += cell
                s += " & "
            s = s[0:-2]
            s += "\\\\ \hline"
            lines.append(s)
        lines.append("\t\\end{tabular}".expandtabs(tabLength))
        if self.__caption is not None:
            lines.append(("\t\caption{" + self.__caption + "}").expandtabs(tabLength))
        lines.append("\\end{table}")
        return lines

    def print(self, tabLength: int = 4, separator: chr = '.'):
        for line in self.lines(tabLength, separator):
            print(line)



from abc import ABC, abstractmethod
import numpy as np


class Formatter(ABC):

    def __init__(self, precision: int, errorPrecision: int):
        self._precision = precision
        self._errorPrecision = errorPrecision

    @abstractmethod
    def format(self, value: float, error: float | None = None):
        pass


class FloatFormatter(Formatter):

    def __init__(self, precision: int = 3, errorPrecision: int = 4):
        super(FloatFormatter, self).__init__(precision, errorPrecision)

    def format(self, value: float, error: float | None = None):
        if error is None:
            return ("{:." + str(self._precision) + "f} ").format(value)
        else:
            template = ("${:." + str(self._precision) + "f} \pm " + "{:." + str(self._errorPrecision) + "f}" + "$ ")
            return template.format(value, error)


class ExponentialFormatter(Formatter):

    def __init__(self, precision: int = 3, errorPrecision: int = 4):
        super(ExponentialFormatter, self).__init__(precision, errorPrecision)

    def format(self, value: float, error: float | None = None):
        if error is None:
            a = int(np.floor(np.log10(np.abs(value))))
            b = value / 10**a
            s = "${:." + str(self._precision) + "f} \\cdot 10^"
            s = s.format(b)
            s += "{" + str(a) + "}$ "
        else:
            a = int(np.floor(np.log10(np.abs(value))))
            b = value / 10**a
            c = error / 10**a
            s = "$({:." + str(self._precision) + "f} \\pm {:." + str(self._errorPrecision) + "f})\\cdot 10^"
            s = s.format(b, c)
            s += "{" + str(a) + "}$ "
        return s
